@@ -8,6 +8,7 @@ from packaging.version import Version
 
 import dask
 from dask import delayed
+from dask._compatibility import WINDOWS
 from dask.base import collections_to_dsk, key_split, visualize_dsk
 from dask.core import get_deps
 from dask.order import _connecting_to_roots, diagnostics, ndependencies, order
@@ -254,6 +255,7 @@ def _rechunk_merge_graph():
     }
 
 
+@pytest.mark.skipif(WINDOWS, reason="Broken pipe on Windows")
 def test_visualize_int_overflow():
     pytest.importorskip("graphviz")
     pytest.importorskip("matplotlib")
@@ -1085,7 +1087,7 @@ def test_array_store_final_order(tmpdir):
     if Version(zarr.__version__) < Version("3.0.0.a0"):
         store = zarr.storage.DirectoryStore(tmpdir)
     else:
-        store = zarr.storage.LocalStore(str(tmpdir), mode="w")
+        store = zarr.storage.LocalStore(str(tmpdir), read_only=False)
 
     root = zarr.group(store, overwrite=True)
     dest = root.empty_like(name="dest", data=x, chunks=x.chunksize, overwrite=True)
@@ -1390,7 +1392,7 @@ def test_xarray_like_reduction():
 )
 def test_array_vs_dataframe(optimize):
     xr = pytest.importorskip("xarray")
-    dd = pytest.importorskip("dask.dataframe")
+    pytest.importorskip("dask.dataframe")
 
     import dask.array as da
 
@@ -1415,7 +1417,7 @@ def test_array_vs_dataframe(optimize):
     diag_df = diagnostics(
         collections_to_dsk([mean.to_dask_dataframe()], optimize_graph=optimize)
     )
-    assert max(diag_df[1]) == (15 if dd._dask_expr_enabled() else 38)
+    assert max(diag_df[1]) == 15
     assert max(diag_array[1]) == 38
     assert max(diag_array[1]) < 50
 
